/*
FreeIMU.h - Interface a FreeIMU device with Arduino
Copyright (C) 2011 Fabio Varesano <fabio at varesano dot net>


This program is free software: you can redistribute it and/or modify
it under the terms of the version 3 GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

//#define DEBUG
#include "WProgram.h"
#include "FreeIMU1.h"
#include "WireUtils.h"
#include "DebugUtils.h"

//----------------------------------------------------------------------------------------------------
// Definitions

//#define Kp 2.0f   // proportional gain governs rate of convergence to accelerometer/magnetometer
//#define Ki 0.005f   // integral gain governs rate of convergence of gyroscope biases
#define halfT 0.02f   // half the sample period


FreeIMU::FreeIMU() {
  //acc = ADXL345();
  gyro = ITG3200();
  //magn = HMC5843();
  
  // initialize quaternion
  q0 = 1;
  q1 = 0;
  q2 = 0;
  q3 = 0;
  //exInt = 0;
  //eyInt = 0;
  //ezInt = 0;
}

void FreeIMU::init() {
  // use default addresses
  init(FIMU_ITG3200_DEF_ADDR);
}

void FreeIMU::init(int gyro_addr) {
  delay(5);
  
  // init ADXL345
  //acc.init(acc_addr);
  
  gyro.init(gyro_addr);
  delay(100);
  /* assuming the device is initialized steady, we do 128 measurements to compute offsets */
  
  float cal_gyro_sum[3];
  float cal_gyro[3];
  float gyro_offsets[3];
  for(int i=0; i<128; i++) {
    gyro.readGyro((double *) cal_gyro);
    cal_gyro_sum[0] += cal_gyro[0];
    cal_gyro_sum[1] += cal_gyro[1];
    cal_gyro_sum[2] += cal_gyro[2];
  }
  gyro_offsets[0] = cal_gyro_sum[0] / 128.0;
  gyro_offsets[1] = cal_gyro_sum[1] / 128.0;
  gyro_offsets[2] = cal_gyro_sum[2] / 128.0;
  gyro.setOffsets((double) -gyro_offsets[0], (double) -gyro_offsets[1], (double) -gyro_offsets[2]);
  
  
  // init HMC5843
  //magn.init(magn_addr, false); // Don't set mode yet, we'll do that later on.
  // Calibrate HMC using self test, not recommended to change the gain after calibration.
  //magn.calibrate(1); // Use gain 1=default, valid 0-7, 7 not recommended.
  // Single mode conversion was used in calibration, now set continuous mode
  //magn.setMode(0);
  //delay(10);
  //magn.setDOR(B110);
}

void FreeIMU::getRawValues(int * raw_values) {
  //acc.readAccel(&raw_values[0], &raw_values[1], &raw_values[2]);
  gyro.readGyroRaw(&raw_values[3], &raw_values[4], &raw_values[5]);
  //magn.getValues(&raw_values[6], &raw_values[7], &raw_values[8]);;
}

void FreeIMU::getValues(float * values) {
  //acc.get_Gxyz(&values[0]);
  
  //int accval[3];
  //acc.readAccel(&accval[0], &accval[1], &accval[2]);
  //values[0] = ((float) accval[0]);
  //values[1] = ((float) accval[1]);
  //values[2] = ((float) accval[2]);
  //delay(5);
  gyro.readGyro((double *) &values[3]);
  DEBUG_PRINT(int(values[3]));
  delay(5);
  //magn.getValues(&values[6]);
  //delay(5);
}


//=====================================================================================================
// AHRS.c
// S.O.H. Madgwick
// 25th August 2010
//=====================================================================================================
// Description:
//
// Quaternion implementation of the 'DCM filter' [Mayhony et al].  Incorporates the magnetic distortion
// compensation algorithms from my filter [Madgwick] which eliminates the need for a reference
// direction of flux (bx bz) to be predefined and limits the effect of magnetic distortions to yaw
// axis only.
//
// User must define 'halfT' as the (sample period / 2), and the filter gains 'Kp' and 'Ki'.
//
// Global variables 'q0', 'q1', 'q2', 'q3' are the quaternion elements representing the estimated
// orientation.  See my report for an overview of the use of quaternions in this application.
//
// User must call 'AHRSupdate()' every sample period and parse calibrated gyroscope ('gx', 'gy', 'gz'),
// accelerometer ('ax', 'ay', 'ay') and magnetometer ('mx', 'my', 'mz') data.  Gyroscope units are
// radians/second, accelerometer and magnetometer units are irrelevant as the vector is normalised.
//
//=====================================================================================================
void FreeIMU::AHRSupdate(float gx, float gy, float gz, float ax, float ay, float az, float mx, float my, float mz) {
  float norm;
  //float hx, hy, hz, bx, bz;
  //float vx, vy, vz, wx, wy, wz;
  //float ex, ey, ez;

  // auxiliary variables to reduce number of repeated operations
  //float q0q0 = q0*q0;
  //float q0q1 = q0*q1;
  //float q0q2 = q0*q2;
  //float q0q3 = q0*q3;
  //float q1q1 = q1*q1;
  //float q1q2 = q1*q2;
  //float q1q3 = q1*q3;
  //float q2q2 = q2*q2;   
  //float q2q3 = q2*q3;
  //float q3q3 = q3*q3;          
  
  // normalise the measurements
  //norm = sqrt(ax*ax + ay*ay + az*az);       
  //ax = ax / norm;
  //ay = ay / norm;
  //az = az / norm;
  //norm = sqrt(mx*mx + my*my + mz*mz);          
  //mx = mx / norm;
  //my = my / norm;
  //mz = mz / norm;         
  
  // compute reference direction of flux
  //hx = 2*mx*(0.5 - q2q2 - q3q3) + 2*my*(q1q2 - q0q3) + 2*mz*(q1q3 + q0q2);
  //hy = 2*mx*(q1q2 + q0q3) + 2*my*(0.5 - q1q1 - q3q3) + 2*mz*(q2q3 - q0q1);
  //hz = 2*mx*(q1q3 - q0q2) + 2*my*(q2q3 + q0q1) + 2*mz*(0.5 - q1q1 - q2q2);         
  //bx = sqrt((hx*hx) + (hy*hy));
  //bz = hz;        
  
  // estimated direction of gravity and flux (v and w)
  //vx = 2*(q1q3 - q0q2);
  //vy = 2*(q0q1 + q2q3);
  //vz = q0q0 - q1q1 - q2q2 + q3q3;
  //wx = 2*bx*(0.5 - q2q2 - q3q3) + 2*bz*(q1q3 - q0q2);
  //wy = 2*bx*(q1q2 - q0q3) + 2*bz*(q0q1 + q2q3);
  //wz = 2*bx*(q0q2 + q1q3) + 2*bz*(0.5 - q1q1 - q2q2);  
  
  // error is sum of cross product between reference direction of fields and direction measured by sensors
  //ex = (ay*vz - az*vy) + (my*wz - mz*wy);
  //ey = (az*vx - ax*vz) + (mz*wx - mx*wz);
  //ez = (ax*vy - ay*vx) + (mx*wy - my*wx);
  
  // integral error scaled integral gain
  //exInt = exInt + ex*Ki;
  //eyInt = eyInt + ey*Ki;
  //ezInt = ezInt + ez*Ki;
  
  // adjusted gyroscope measurements
  //gx = gx + Kp*ex + exInt;
  //gy = gy + Kp*ey + eyInt;
  //gz = gz + Kp*ez + ezInt;
  
  // integrate quaternion rate and normalise
  if(abs(gx)<0.03 && abs(gy)<0.03 && abs(gz)<0.03){
    q0 = q0; q1 = q1; q2 = q2; q3 = q3;
  }else{  
    q0 = q0 + (-q1*gx - q2*gy - q3*gz)*halfT;
    q1 = q1 + (q0*gx + q2*gz - q3*gy)*halfT;
    q2 = q2 + (q0*gy - q1*gz + q3*gx)*halfT;
    q3 = q3 + (q0*gz + q1*gy - q2*gx)*halfT;  
  }  

  // normalise quaternion
  norm = sqrt(q0*q0 + q1*q1 + q2*q2 + q3*q3);
  q0 = q0 / norm;
  q1 = q1 / norm;
  q2 = q2 / norm;
  q3 = q3 / norm;
}

void FreeIMU::getQ(float * q) {
  float val[9];
  getValues(val);
  
  DEBUG_PRINT(val[3] * M_PI/180);
  DEBUG_PRINT(val[4] * M_PI/180);
  DEBUG_PRINT(val[5] * M_PI/180);
  DEBUG_PRINT(val[0]);
  DEBUG_PRINT(val[1]);
  DEBUG_PRINT(val[2]);
  DEBUG_PRINT(val[6]);
  DEBUG_PRINT(val[7]);
  DEBUG_PRINT(val[8]);
  
  AHRSupdate(val[3] * M_PI/180, val[4] * M_PI/180, val[5] * M_PI/180, val[0], val[1], val[2], val[6], val[7], val[8]);
  q[0] = q0;
  q[1] = q1;
  q[2] = q2;
  q[3] = q3;
}

void FreeIMU::SetInitQ(float init_q0){
  q0 = init_q0;
  q1 = 0;
  q2 = 0;
  q3 = sqrt(1 - q0*q0);
}


